@@ -10,6 +10,7 @@
  *               Arnaldo Carvalho de Melo <acme@conectiva.com.br>, Mar 1999,
  *               Phillip Kesling <pkesling@sgi.com>, Mar 2003.
  */
+
 #include "c.h"
 #include "nls.h"
 #include "all-io.h"
@@ -21,6 +22,13 @@
 #include "pt-mbr.h"
 #include "fdiskP.h"
 
+/**
+ * SECTION: sgi
+ * @title: SGI
+ * @short_description: disk label specific functions
+ *
+ */
+
 /*
  * in-memory fdisk SGI stuff
  */
@@ -67,7 +75,7 @@ static inline struct sgi_disklabel *self_disklabel(struct fdisk_context *cxt)
 {
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
 	return ((struct fdisk_sgi_label *) cxt->label)->header;
 }
@@ -77,7 +85,7 @@ static inline struct fdisk_sgi_label *self_label(struct fdisk_context *cxt)
 {
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
 	return (struct fdisk_sgi_label *) cxt->label;
 }
@@ -129,6 +137,16 @@ static void sgi_free_info(struct sgi_info *info)
 	free(info);
 }
 
+/**
+ * fdisk_sgi_create_info:
+ * @cxt: context
+ *
+ * This function add hint about SGI label (e.g. set "sgilabel" as volume name)
+ * to the first SGI volume. This is probably old SGI convention without any
+ * effect to the device partitioning.
+ *
+ * Returns: 0 on success, <0 on error
+ */
 int fdisk_sgi_create_info(struct fdisk_context *cxt)
 {
 	struct sgi_disklabel *sgilabel = self_disklabel(cxt);
@@ -138,7 +156,7 @@ int fdisk_sgi_create_info(struct fdisk_context *cxt)
 	sgilabel->volume[0].num_bytes = cpu_to_be32(sizeof(struct sgi_info));
 	strncpy((char *) sgilabel->volume[0].name, "sgilabel", 8);
 
-	fdisk_info(cxt, _("SGI info created on second sector"));
+	fdisk_info(cxt, _("SGI info created on second sector."));
 	return 0;
 }
 
@@ -221,7 +239,7 @@ static int sgi_probe_label(struct fdisk_context *cxt)
 
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 	assert(sizeof(struct sgi_disklabel) <= 512);
 
 	/* map first sector to header */
@@ -246,124 +264,48 @@ static int sgi_probe_label(struct fdisk_context *cxt)
 	return 1;
 }
 
-static int sgi_list_table(struct fdisk_context *cxt)
+static int sgi_get_disklabel_item(struct fdisk_context *cxt, struct fdisk_labelitem *item)
 {
-	struct tt *tb = NULL;
-	struct sgi_disklabel *sgilabel = self_disklabel(cxt);
-	struct sgi_device_parameter *sgiparam = &sgilabel->devparam;
-	size_t i, used;
-	char *p;
-	int rc;
+	struct sgi_disklabel *sgilabel;
+	struct sgi_device_parameter *sgiparam;
+	int rc = 0;
 
-	if (fdisk_context_display_details(cxt))
-		fdisk_colon(cxt, _(
-			"Label geometry: %d heads, %llu sectors\n"
-			"                %llu cylinders, %d physical cylinders\n"
-			"                %d extra sects/cyl, interleave %d:1\n"),
-			cxt->geom.heads, cxt->geom.sectors,
-			cxt->geom.cylinders, be16_to_cpu(sgiparam->pcylcount),
-			(int) sgiparam->sparecyl, be16_to_cpu(sgiparam->ilfact));
+	assert(cxt);
+	assert(cxt->label);
+	assert(fdisk_is_label(cxt, SGI));
 
-	/*
-	 * Partitions
-	 */
-	tb = tt_new_table(TT_FL_FREEDATA);
-	if (!tb)
-		return -ENOMEM;
+	sgilabel = self_disklabel(cxt);
+	sgiparam = &sgilabel->devparam;
 
-	tt_define_column(tb, _("Pt#"),      3, TT_FL_RIGHT);
-	tt_define_column(tb, _("Device"), 0.2, 0);
-	tt_define_column(tb, _("Info"),     2, 0);
-	tt_define_column(tb, _("Start"),    9, TT_FL_RIGHT);
-	tt_define_column(tb, _("End"),      9, TT_FL_RIGHT);
-	tt_define_column(tb, _("Sectors"),  9, TT_FL_RIGHT);
-	tt_define_column(tb, _("Id"),       2, TT_FL_RIGHT);
-	tt_define_column(tb, _("System"), 0.2, TT_FL_TRUNC);
-
-	for (i = 0, used = 1; i < cxt->label->nparts_max; i++) {
-		uint32_t start, len;
-		struct fdisk_parttype *t;
-		struct tt_line *ln;
-
-		if (sgi_get_num_sectors(cxt, i) == 0)
-			continue;
-
-		ln = tt_add_line(tb, NULL);
-		if (!ln)
-			continue;
-		start = sgi_get_start_sector(cxt, i);
-		len = sgi_get_num_sectors(cxt, i);
-		t = fdisk_get_partition_type(cxt, i);
-
-		if (asprintf(&p, "%zu:", i + 1) > 0)
-			tt_line_set_data(ln, 0, p);	/* # */
-		p = fdisk_partname(cxt->dev_path, used++);
-		if (p)
-			tt_line_set_data(ln, 1, p);	/* Device */
-
-		p = sgi_get_swappartition(cxt) == (int) i ? "swap" :
-		    sgi_get_bootpartition(cxt) == (int) i ? "boot" : NULL;
-		if (p)
-			tt_line_set_data(ln, 2, strdup(p));	/* Info */
-
-		if (asprintf(&p, "%ju", (uintmax_t) fdisk_scround(cxt, start)) > 0)
-			tt_line_set_data(ln, 3, p);	/* Start */
-		if (asprintf(&p, "%ju",	(uintmax_t) fdisk_scround(cxt, start + len) - 1) > 0)
-			tt_line_set_data(ln, 4, p);	/* End */
-		if (asprintf(&p, "%ju",	(uintmax_t) len) > 0)
-			tt_line_set_data(ln, 5, p);	/* Sectors*/
-		if (asprintf(&p, "%2x", t->type) > 0)
-			tt_line_set_data(ln, 6, p);	/* type ID */
-		if (t->name)
-			tt_line_set_data(ln, 7, strdup(t->name)); /* type Name */
-		fdisk_free_parttype(t);
+	switch (item->id) {
+	case SGI_LABELITEM_PCYLCOUNT:
+		item->name = _("Physical cylinders");
+		item->type = 'j';
+		item->data.num64 = (uint64_t) be16_to_cpu(sgiparam->pcylcount);
+		break;
+	case SGI_LABELITEM_SPARECYL:
+		item->name = _("Extra sects/cyl");
+		item->type = 'j';
+		item->data.num64 = (uint64_t) sgiparam->sparecyl;
+		break;
+	case SGI_LABELITEM_ILFACT:
+		item->name = _("Interleave");
+		item->type = 'j';
+		item->data.num64 = (uint64_t) be16_to_cpu(sgiparam->ilfact);
+		break;
+	case SGI_LABELITEM_BOOTFILE:
+		item->name = _("Bootfile");
+		item->type = 's';
+		item->data.str = *sgilabel->boot_file ? strdup((char *) sgilabel->boot_file) : NULL;
+		break;
+	default:
+		if (item->id < __FDISK_NLABELITEMS)
+			rc = 1;	/* unssupported generic item */
+		else
+			rc = 2;	/* out of range */
+		break;
 	}
 
-	rc = fdisk_print_table(cxt, tb);
-	tt_free_table(tb);
-	if (rc)
-		return rc;
-
-	/*
-	 * Volumes
-	 */
-	tb = tt_new_table(TT_FL_FREEDATA);
-	if (!tb)
-		return -ENOMEM;
-
-	tt_define_column(tb, _("#"),       3, TT_FL_RIGHT);
-	tt_define_column(tb, _("Name"),  0.2, 0);
-	tt_define_column(tb, _("Sector"),  2, TT_FL_RIGHT);
-	tt_define_column(tb, _("Size"),    9, TT_FL_RIGHT);
-
-	for (i = 0, used = 0; i < SGI_MAXVOLUMES; i++) {
-		struct tt_line *ln;
-		uint32_t start = be32_to_cpu(sgilabel->volume[i].block_num),
-			 len = be32_to_cpu(sgilabel->volume[i].num_bytes);
-		if (!len)
-			continue;
-		ln = tt_add_line(tb, NULL);
-		if (!ln)
-			continue;
-		if (asprintf(&p, "%zu:", i) > 0)
-			tt_line_set_data(ln, 0, p);		/* # */
-		if (*sgilabel->volume[i].name)
-			tt_line_set_data(ln, 1,
-				strndup((char *) sgilabel->volume[i].name,
-				        sizeof(sgilabel->volume[i].name)));	/* Name */
-		if (asprintf(&p, "%ju", (uintmax_t) start) > 0)
-			tt_line_set_data(ln, 2, p);	/* Sector */
-		if (asprintf(&p, "%ju",	(uintmax_t) len) > 0)
-			tt_line_set_data(ln, 3, p);	/* Size */
-		used++;
-	}
-
-	if (used)
-		rc = fdisk_print_table(cxt, tb);
-	tt_free_table(tb);
-
-	fdisk_colon(cxt, _("Bootfile: %s"), sgilabel->boot_file);
-
 	return rc;
 }
 
@@ -402,6 +344,45 @@ static unsigned int sgi_get_lastblock(struct fdisk_context *cxt)
 	return cxt->geom.heads * cxt->geom.sectors * cxt->geom.cylinders;
 }
 
+static struct fdisk_parttype *sgi_get_parttype(struct fdisk_context *cxt, size_t n)
+{
+	struct fdisk_parttype *t;
+
+	if (n >= cxt->label->nparts_max)
+		return NULL;
+
+	t = fdisk_label_get_parttype_from_code(cxt->label, sgi_get_sysid(cxt, n));
+	return t ? : fdisk_new_unknown_parttype(sgi_get_sysid(cxt, n), NULL);
+}
+
+/* fdisk_get_partition() backend */
+static int sgi_get_partition(struct fdisk_context *cxt, size_t n, struct fdisk_partition *pa)
+{
+	fdisk_sector_t start, len;
+
+	pa->used = sgi_get_num_sectors(cxt, n) > 0;
+	if (!pa->used)
+		return 0;
+
+	start = sgi_get_start_sector(cxt, n);
+	len = sgi_get_num_sectors(cxt, n);
+
+	pa->type = sgi_get_parttype(cxt, n);
+	pa->size = len;
+	pa->start = start;
+
+	if (pa->type && pa->type->code == SGI_TYPE_ENTIRE_DISK)
+		pa->wholedisk = 1;
+
+	pa->attrs = sgi_get_swappartition(cxt) == (int) n ? "swap" :
+		    sgi_get_bootpartition(cxt) == (int) n ? "boot" : NULL;
+	if (pa->attrs)
+		pa->attrs = strdup(pa->attrs);
+
+	return 0;
+}
+
+
 static int sgi_check_bootfile(struct fdisk_context *cxt, const char *name)
 {
 	size_t sz;
@@ -439,6 +420,15 @@ static int sgi_check_bootfile(struct fdisk_context *cxt, const char *name)
 	return 1;	/* filename did not change */
 }
 
+/**
+ * fdisk_sgi_set_bootfile:
+ * @cxt: context
+ *
+ * Allows to set SGI boot file. The function uses Ask API for dialog with
+ * user.
+ *
+ * Returns: 0 on success, <0 on error
+ */
 int fdisk_sgi_set_bootfile(struct fdisk_context *cxt)
 {
 	int rc = 0;
@@ -464,8 +454,7 @@ int fdisk_sgi_set_bootfile(struct fdisk_context *cxt)
 
 	memcpy(sgilabel->boot_file, name, sz);
 
-	fdisk_sinfo(cxt, FDISK_INFO_SUCCESS,
-			_("Bootfile has been changed to \"%s\"."), name);
+	fdisk_info(cxt, _("Bootfile has been changed to \"%s\"."), name);
 done:
 	free(name);
 	return rc;
@@ -478,7 +467,7 @@ static int sgi_write_disklabel(struct fdisk_context *cxt)
 
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
 	sgilabel = self_disklabel(cxt);
 	sgilabel->csum = 0;
@@ -594,7 +583,7 @@ static int verify_disklabel(struct fdisk_context *cxt, int verbose)
 
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
 	clear_freelist(cxt);
 	memset(Index, 0, sizeof(Index));
@@ -613,6 +602,8 @@ static int verify_disklabel(struct fdisk_context *cxt, int verbose)
 	if (sortcount == 0) {
 		if (verbose)
 			fdisk_info(cxt, _("No partitions defined."));
+		if (lastblock)
+			add_to_freelist(cxt, 0, lastblock);
 		return (lastblock > 0) ? 1 : (lastblock == 0) ? 0 : -1;
 	}
 
@@ -629,14 +620,14 @@ static int verify_disklabel(struct fdisk_context *cxt, int verbose)
 				   sgi_get_start_sector(cxt, Index[0]));
 
 		if (verbose && sgi_get_num_sectors(cxt, Index[0]) != lastblock)
-			DBG(LABEL, dbgprint(
+			DBG(LABEL, ul_debug(
 				"entire disk partition=%ds, but disk=%ds",
 				sgi_get_num_sectors(cxt, Index[0]),
 				lastblock));
 		lastblock = sgi_get_num_sectors(cxt, Index[0]);
 	} else if (verbose) {
 		fdisk_info(cxt, _("Partition 11 should cover the entire disk."));
-		DBG(LABEL, dbgprint("sysid=%d\tpartition=%d",
+		DBG(LABEL, ul_debug("sysid=%d\tpartition=%d",
 			       sgi_get_sysid(cxt, Index[0]), Index[0]+1));
 	}
 	for (i=1, start=0; i<sortcount; i++) {
@@ -644,12 +635,12 @@ static int verify_disklabel(struct fdisk_context *cxt, int verbose)
 
 		if (verbose && cylsize
 		    && (sgi_get_start_sector(cxt, Index[i]) % cylsize) != 0)
-			DBG(LABEL, dbgprint("partition %d does not start on "
+			DBG(LABEL, ul_debug("partition %d does not start on "
 					"cylinder boundary.", Index[i]+1));
 
 		if (verbose && cylsize
 		    && sgi_get_num_sectors(cxt, Index[i]) % cylsize != 0)
-			DBG(LABEL, dbgprint("partition %d does not end on "
+			DBG(LABEL, ul_debug("partition %d does not end on "
 					"cylinder boundary.", Index[i]+1));
 
 		/* We cannot handle several "entire disk" entries. */
@@ -685,7 +676,7 @@ static int verify_disklabel(struct fdisk_context *cxt, int verbose)
 		if (cylsize && start % cylsize)
 			start += cylsize - (start % cylsize);
 
-		DBG(LABEL, dbgprint("%2d:%12d\t%12d\t%12d", Index[i],
+		DBG(LABEL, ul_debug("%2d:%12d\t%12d\t%12d", Index[i],
 				       sgi_get_start_sector(cxt, Index[i]),
 				       sgi_get_num_sectors(cxt, Index[i]),
 				       sgi_get_sysid(cxt, Index[i])));
@@ -749,14 +740,14 @@ static int sgi_entire(struct fdisk_context *cxt)
 	return -1;
 }
 
-static int sgi_set_partition(struct fdisk_context *cxt, size_t i,
+static int set_partition(struct fdisk_context *cxt, size_t i,
 			     unsigned int start, unsigned int length, int sys)
 {
 	struct sgi_disklabel *sgilabel;
 
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
 	sgilabel = self_disklabel(cxt);
 	sgilabel->partitions[i].type = cpu_to_be32(sys);
@@ -768,7 +759,8 @@ static int sgi_set_partition(struct fdisk_context *cxt, size_t i,
 	if (sgi_gaps(cxt) < 0)	/* rebuild freelist */
 		fdisk_warnx(cxt, _("Partition overlap on the disk."));
 	if (length) {
-		struct fdisk_parttype *t = fdisk_get_parttype_from_code(cxt, sys);
+		struct fdisk_parttype *t =
+				fdisk_label_get_parttype_from_code(cxt->label, sys);
 		fdisk_info_new_partition(cxt, i + 1, start, start + length, t);
 	}
 
@@ -781,7 +773,7 @@ static void sgi_set_entire(struct fdisk_context *cxt)
 
 	for (n = 10; n < cxt->label->nparts_max; n++) {
 		if (!sgi_get_num_sectors(cxt, n)) {
-			sgi_set_partition(cxt, n, 0, sgi_get_lastblock(cxt), SGI_TYPE_ENTIRE_DISK);
+			set_partition(cxt, n, 0, sgi_get_lastblock(cxt), SGI_TYPE_ENTIRE_DISK);
 			break;
 		}
 	}
@@ -795,7 +787,7 @@ static void sgi_set_volhdr(struct fdisk_context *cxt)
 		if (!sgi_get_num_sectors(cxt, n)) {
 			/* Choose same default volume header size as IRIX fx uses. */
 			if (4096 < sgi_get_lastblock(cxt))
-				sgi_set_partition(cxt, n, 0, 4096, SGI_TYPE_VOLHDR);
+				set_partition(cxt, n, 0, 4096, SGI_TYPE_VOLHDR);
 			break;
 		}
 	}
@@ -811,7 +803,7 @@ static int sgi_delete_partition(struct fdisk_context *cxt, size_t partnum)
 	if (partnum > cxt->label->nparts_max)
 		return -EINVAL;
 
-	rc = sgi_set_partition(cxt, partnum, 0, 0, 0);
+	rc = set_partition(cxt, partnum, 0, 0, 0);
 
 	cxt->label->nparts_cur = count_used_partitions(cxt);
 
@@ -819,20 +811,24 @@ static int sgi_delete_partition(struct fdisk_context *cxt, size_t partnum)
 }
 
 static int sgi_add_partition(struct fdisk_context *cxt,
-		size_t n,
-		struct fdisk_parttype *t)
+			     struct fdisk_partition *pa,
+			     size_t *partno)
 {
 	struct fdisk_sgi_label *sgi;
 	char mesg[256];
 	unsigned int first = 0, last = 0;
 	struct fdisk_ask *ask;
-	int sys = t ? t->type : SGI_TYPE_XFS;
+	int sys = pa && pa->type ? pa->type->code : SGI_TYPE_XFS;
 	int rc;
+	size_t n;
 
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
+	rc = fdisk_partition_next_partno(pa, cxt, &n);
+	if (rc)
+		return rc;
 	if (n == 10)
 		sys = SGI_TYPE_ENTIRE_DISK;
 	else if (n == 8)
@@ -841,11 +837,11 @@ static int sgi_add_partition(struct fdisk_context *cxt,
 	sgi = self_label(cxt);
 
 	if (sgi_get_num_sectors(cxt, n)) {
-		fdisk_warnx(cxt, _("Partition %zd is already defined.  "
+		fdisk_warnx(cxt, _("Partition %zu is already defined.  "
 				   "Delete it before re-adding it."), n + 1);
 		return -EINVAL;
 	}
-	if (sgi_entire(cxt) == -1 &&  sys != SGI_TYPE_ENTIRE_DISK) {
+	if (!cxt->script && sgi_entire(cxt) == -1 &&  sys != SGI_TYPE_ENTIRE_DISK) {
 		fdisk_info(cxt, _("Attempting to generate entire disk entry automatically."));
 		sgi_set_entire(cxt);
 		sgi_set_volhdr(cxt);
@@ -859,9 +855,26 @@ static int sgi_add_partition(struct fdisk_context *cxt,
 		return -EINVAL;
 	}
 
-	snprintf(mesg, sizeof(mesg), _("First %s"),
-			fdisk_context_get_unit(cxt, SINGULAR));
-	for (;;) {
+	if (sys == SGI_TYPE_ENTIRE_DISK) {
+		first = 0;
+		last = sgi_get_lastblock(cxt);
+	} else {
+		first = sgi->freelist[0].first;
+		last  = sgi->freelist[0].last;
+	}
+
+	/* first sector */
+	if (pa && pa->start_follow_default)
+		;
+	else if (pa && fdisk_partition_has_start(pa)) {
+		first = pa->start;
+		last = is_in_freelist(cxt, first);
+
+		if (sys != SGI_TYPE_ENTIRE_DISK && !last)
+			return -ERANGE;
+	} else {
+		snprintf(mesg, sizeof(mesg), _("First %s"),
+				fdisk_get_unit(cxt, FDISK_SINGULAR));
 		ask = fdisk_new_ask();
 		if (!ask)
 			return -ENOMEM;
@@ -869,76 +882,68 @@ static int sgi_add_partition(struct fdisk_context *cxt,
 		fdisk_ask_set_query(ask, mesg);
 		fdisk_ask_set_type(ask, FDISK_ASKTYPE_NUMBER);
 
-		if (sys == SGI_TYPE_ENTIRE_DISK) {
-			last = sgi_get_lastblock(cxt);
-			fdisk_ask_number_set_low(ask,     0);	/* minimal */
-			fdisk_ask_number_set_default(ask, 0);	/* default */
-			fdisk_ask_number_set_high(ask, last - 1); /* maximal */
-		} else {
-			first = sgi->freelist[0].first;
-			last  = sgi->freelist[0].last;
-			fdisk_ask_number_set_low(ask,     fdisk_scround(cxt, first));	/* minimal */
-			fdisk_ask_number_set_default(ask, fdisk_scround(cxt, first));	/* default */
-			fdisk_ask_number_set_high(ask,    fdisk_scround(cxt, last) - 1); /* maximal */
-		}
+		fdisk_ask_number_set_low(ask,     fdisk_scround(cxt, first));	/* minimal */
+		fdisk_ask_number_set_default(ask, fdisk_scround(cxt, first));	/* default */
+		fdisk_ask_number_set_high(ask,    fdisk_scround(cxt, last) - 1); /* maximal */
+
 		rc = fdisk_do_ask(cxt, ask);
 		first = fdisk_ask_number_get_result(ask);
-		fdisk_free_ask(ask);
+		fdisk_unref_ask(ask);
 
 		if (rc)
 			return rc;
-
-		if (first && sys == SGI_TYPE_ENTIRE_DISK)
-			fdisk_info(cxt, _("It is highly recommended that the "
-					  "eleventh partition covers the entire "
-					  "disk and is of type 'SGI volume'."));
-
-		if (fdisk_context_use_cylinders(cxt))
-			first *= fdisk_context_get_units_per_sector(cxt);
-		/*else
-			first = first; * align to cylinder if you know how ... */
-		if (!last)
-			last = is_in_freelist(cxt, first);
-		if (last == 0)
-			fdisk_warnx(cxt, _("You will get a partition overlap "
-					   "on the disk. Fix it first!"));
-		else
-			break;
+		if (fdisk_use_cylinders(cxt))
+			first *= fdisk_get_units_per_sector(cxt);
 	}
 
-	snprintf(mesg, sizeof(mesg),
-		 _("Last %s or +%s or +size{K,M,G,T,P}"),
-		 fdisk_context_get_unit(cxt, SINGULAR),
-		 fdisk_context_get_unit(cxt, PLURAL));
+	if (first && sys == SGI_TYPE_ENTIRE_DISK)
+		fdisk_info(cxt, _("It is highly recommended that the "
+				  "eleventh partition covers the entire "
+				  "disk and is of type 'SGI volume'."));
+	if (!last)
+		last = is_in_freelist(cxt, first);
 
-	ask = fdisk_new_ask();
-	if (!ask)
-		return -ENOMEM;
+	/* last sector */
+	if (pa && pa->end_follow_default)
+		last -= 1ULL;
+	else if (pa && fdisk_partition_has_size(pa)) {
+		if (first + pa->size - 1ULL > last)
+			return -ERANGE;
+		last = first + pa->size - 1ULL;
+	} else {
+		snprintf(mesg, sizeof(mesg),
+			 _("Last %s or +%s or +size{K,M,G,T,P}"),
+			 fdisk_get_unit(cxt, FDISK_SINGULAR),
+			 fdisk_get_unit(cxt, FDISK_PLURAL));
 
-	fdisk_ask_set_query(ask, mesg);
-	fdisk_ask_set_type(ask, FDISK_ASKTYPE_OFFSET);
+		ask = fdisk_new_ask();
+		if (!ask)
+			return -ENOMEM;
 
-	fdisk_ask_number_set_low(ask,     fdisk_scround(cxt, first));	/* minimal */
-	fdisk_ask_number_set_default(ask, fdisk_scround(cxt, last) - 1);/* default */
-	fdisk_ask_number_set_high(ask,    fdisk_scround(cxt, last) - 1);/* maximal */
-	fdisk_ask_number_set_base(ask,    fdisk_scround(cxt, first));
+		fdisk_ask_set_query(ask, mesg);
+		fdisk_ask_set_type(ask, FDISK_ASKTYPE_OFFSET);
 
-	if (fdisk_context_use_cylinders(cxt))
-		fdisk_ask_number_set_unit(ask,
-			     cxt->sector_size *
-			     fdisk_context_get_units_per_sector(cxt));
-	else
-		fdisk_ask_number_set_unit(ask,cxt->sector_size);
+		fdisk_ask_number_set_low(ask,     fdisk_scround(cxt, first));	/* minimal */
+		fdisk_ask_number_set_default(ask, fdisk_scround(cxt, last) - 1);/* default */
+		fdisk_ask_number_set_high(ask,    fdisk_scround(cxt, last) - 1);/* maximal */
+		fdisk_ask_number_set_base(ask,    fdisk_scround(cxt, first));
 
-	rc = fdisk_do_ask(cxt, ask);
-	last = fdisk_ask_number_get_result(ask) + 1;
+		if (fdisk_use_cylinders(cxt))
+			fdisk_ask_number_set_unit(ask,
+				     cxt->sector_size *
+				     fdisk_get_units_per_sector(cxt));
+		else
+			fdisk_ask_number_set_unit(ask,cxt->sector_size);
 
-	fdisk_free_ask(ask);
-	if (rc)
-		return rc;
+		rc = fdisk_do_ask(cxt, ask);
+		last = fdisk_ask_number_get_result(ask) + 1;
 
-	if (fdisk_context_use_cylinders(cxt))
-		last *= fdisk_context_get_units_per_sector(cxt);
+		fdisk_unref_ask(ask);
+		if (rc)
+			return rc;
+		if (fdisk_use_cylinders(cxt))
+			last *= fdisk_get_units_per_sector(cxt);
+	}
 
 	if (sys == SGI_TYPE_ENTIRE_DISK
 	    && (first != 0 || last != sgi_get_lastblock(cxt)))
@@ -946,9 +951,10 @@ static int sgi_add_partition(struct fdisk_context *cxt,
 				  "eleventh partition covers the entire "
 				  "disk and is of type 'SGI volume'."));
 
-	sgi_set_partition(cxt, n, first, last - first, sys);
+	set_partition(cxt, n, first, last - first, sys);
 	cxt->label->nparts_cur = count_used_partitions(cxt);
-
+	if (partno)
+		*partno = n;
 	return 0;
 }
 
@@ -956,18 +962,18 @@ static int sgi_create_disklabel(struct fdisk_context *cxt)
 {
 	struct fdisk_sgi_label *sgi;
 	struct sgi_disklabel *sgilabel;
+	int rc;
 
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
-#ifdef HDIO_GETGEO
 	if (cxt->geom.heads && cxt->geom.sectors) {
-		sector_t llsectors;
+		fdisk_sector_t llsectors;
 
-		if (blkdev_get_sectors(cxt->dev_fd, &llsectors) == 0) {
+		if (blkdev_get_sectors(cxt->dev_fd, (unsigned long long *) &llsectors) == 0) {
 			/* the get device size ioctl was successful */
-			sector_t llcyls;
+			fdisk_sector_t llcyls;
 			int sec_fac = cxt->sector_size / 512;
 
 			llcyls = llsectors / (cxt->geom.heads * cxt->geom.sectors * sec_fac);
@@ -983,8 +989,11 @@ static int sgi_create_disklabel(struct fdisk_context *cxt)
 				  "> 33.8 GB."), cxt->dev_path, cxt->geom.cylinders);
 		}
 	}
-#endif
-	fdisk_zeroize_firstsector(cxt);
+
+	rc = fdisk_init_firstsector_buffer(cxt, 0, 0);
+	if (rc)
+		return rc;
+
 	sgi = (struct fdisk_sgi_label *) cxt->label;
 	sgi->header = (struct sgi_disklabel *) cxt->firstsector;
 
@@ -1032,85 +1041,80 @@ static int sgi_create_disklabel(struct fdisk_context *cxt)
 	memset(&(sgilabel->partitions), 0,
 			sizeof(struct sgi_partition) * SGI_MAXPARTITIONS);
 	cxt->label->nparts_max = SGI_MAXPARTITIONS;
-	sgi_set_entire(cxt);
-	sgi_set_volhdr(cxt);
 
+	/* don't create default layout when a script defined */
+	if (!cxt->script) {
+		sgi_set_entire(cxt);
+		sgi_set_volhdr(cxt);
+	}
 	cxt->label->nparts_cur = count_used_partitions(cxt);
 
-	fdisk_sinfo(cxt, FDISK_INFO_SUCCESS,
-			_("Created a new SGI disklabel."));
+	fdisk_info(cxt, _("Created a new SGI disklabel."));
 	return 0;
 }
 
-static struct fdisk_parttype *sgi_get_parttype(struct fdisk_context *cxt, size_t n)
-{
-	struct fdisk_parttype *t;
-
-	if (n >= cxt->label->nparts_max)
-		return NULL;
-
-	t = fdisk_get_parttype_from_code(cxt, sgi_get_sysid(cxt, n));
-	if (!t)
-		t = fdisk_new_unknown_parttype(sgi_get_sysid(cxt, n), NULL);
-	return t;
-}
-
-static int sgi_set_parttype(struct fdisk_context *cxt,
+static int sgi_set_partition(struct fdisk_context *cxt,
 		size_t i,
-		struct fdisk_parttype *t)
+		struct fdisk_partition *pa)
 {
 	struct sgi_disklabel *sgilabel;
 
-	if (i >= cxt->label->nparts_max || !t || t->type > UINT32_MAX)
+	if (i >= cxt->label->nparts_max)
 		return -EINVAL;
 
-	if (sgi_get_num_sectors(cxt, i) == 0)	/* caught already before, ... */ {
-		fdisk_warnx(cxt, _("Sorry, only for non-empty partitions you can change the tag."));
-		return -EINVAL;
-	}
-
-	if ((i == 10 && t->type != SGI_TYPE_ENTIRE_DISK)
-	    || (i == 8 && t->type != 0))
-		fdisk_info(cxt, _("Consider leaving partition 9 as volume header (0), "
-				  "and partition 11 as entire volume (6), "
-				  "as IRIX expects it."));
-
-	if (((t->type != SGI_TYPE_ENTIRE_DISK) && (t->type != SGI_TYPE_VOLHDR))
-	    && (sgi_get_start_sector(cxt, i) < 1)) {
-		int yes = 0;
-		fdisk_ask_yesno(cxt,
-			_("It is highly recommended that the partition at offset 0 "
-			  "is of type \"SGI volhdr\", the IRIX system will rely on it to "
-			  "retrieve from its directory standalone tools like sash and fx. "
-			  "Only the \"SGI volume\" entire disk section may violate this. "
-			  "Are you sure about tagging this partition differently?"), &yes);
-		if (!yes)
-			return 1;
-	}
-
 	sgilabel = self_disklabel(cxt);
-	sgilabel->partitions[i].type = cpu_to_be32(t->type);
+
+	if (pa->type) {
+		struct fdisk_parttype *t = pa->type;
+
+		if (sgi_get_num_sectors(cxt, i) == 0)	/* caught already before, ... */ {
+			fdisk_warnx(cxt, _("Sorry, only for non-empty partitions you can change the tag."));
+			return -EINVAL;
+		}
+
+		if ((i == 10 && t->code != SGI_TYPE_ENTIRE_DISK)
+		    || (i == 8 && t->code != 0))
+			fdisk_info(cxt, _("Consider leaving partition 9 as volume header (0), "
+					  "and partition 11 as entire volume (6), "
+					  "as IRIX expects it."));
+
+		if (cxt->script == NULL
+		    && ((t->code != SGI_TYPE_ENTIRE_DISK) && (t->code != SGI_TYPE_VOLHDR))
+		    && (sgi_get_start_sector(cxt, i) < 1)) {
+			int yes = 0;
+			fdisk_ask_yesno(cxt,
+				_("It is highly recommended that the partition at offset 0 "
+				  "is of type \"SGI volhdr\", the IRIX system will rely on it to "
+				  "retrieve from its directory standalone tools like sash and fx. "
+				  "Only the \"SGI volume\" entire disk section may violate this. "
+				  "Are you sure about tagging this partition differently?"), &yes);
+			if (!yes)
+				return 1;
+		}
+
+		sgilabel->partitions[i].type = cpu_to_be32(t->code);
+	}
+
+	if (fdisk_partition_has_start(pa))
+		sgilabel->partitions[i].first_block = cpu_to_be32(pa->start);
+	if (fdisk_partition_has_size(pa))
+		sgilabel->partitions[i].num_blocks = cpu_to_be32(pa->size);
+
+	fdisk_label_set_changed(cxt->label, 1);
 	return 0;
 }
 
 
-static int sgi_get_partition_status(
+static int sgi_partition_is_used(
 		struct fdisk_context *cxt,
-		size_t i,
-		int *status)
+		size_t i)
 {
 	assert(cxt);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
-	if (!status || i >= cxt->label->nparts_max)
-		return -EINVAL;
-
-	*status = FDISK_PARTSTAT_NONE;
-
-	if (sgi_get_num_sectors(cxt, i))
-		*status = FDISK_PARTSTAT_USED;
-
-	return 0;
+	if (i >= cxt->label->nparts_max)
+		return 0;
+	return sgi_get_num_sectors(cxt, i) ? 1 : 0;
 }
 
 static int sgi_toggle_partition_flag(struct fdisk_context *cxt, size_t i, unsigned long flag)
@@ -1118,7 +1122,7 @@ static int sgi_toggle_partition_flag(struct fdisk_context *cxt, size_t i, unsign
 	struct sgi_disklabel *sgilabel;
 	assert(cxt);
 	assert(cxt->label);
-	assert(fdisk_is_disklabel(cxt, SGI));
+	assert(fdisk_is_label(cxt, SGI));
 
 	if (i >= cxt->label->nparts_max)
 		return -EINVAL;
@@ -1145,19 +1149,33 @@ static int sgi_toggle_partition_flag(struct fdisk_context *cxt, size_t i, unsign
 	return 0;
 }
 
+static const struct fdisk_field sgi_fields[] =
+{
+	{ FDISK_FIELD_DEVICE,	N_("Device"),	 10,	0 },
+	{ FDISK_FIELD_START,	N_("Start"),	  5,	FDISK_FIELDFL_NUMBER },
+	{ FDISK_FIELD_END,	N_("End"),	  5,	FDISK_FIELDFL_NUMBER },
+	{ FDISK_FIELD_SECTORS,	N_("Sectors"),	  5,	FDISK_FIELDFL_NUMBER },
+	{ FDISK_FIELD_CYLINDERS,N_("Cylinders"),  5,	FDISK_FIELDFL_NUMBER },
+	{ FDISK_FIELD_SIZE,	N_("Size"),	  5,	FDISK_FIELDFL_NUMBER | FDISK_FIELDFL_EYECANDY },
+	{ FDISK_FIELD_TYPEID,	N_("Id"),	  2,	FDISK_FIELDFL_NUMBER },
+	{ FDISK_FIELD_TYPE,	N_("Type"),	0.1,	FDISK_FIELDFL_EYECANDY },
+	{ FDISK_FIELD_ATTR,	N_("Attrs"),	  0,	FDISK_FIELDFL_NUMBER }
+};
+
 static const struct fdisk_label_operations sgi_operations =
 {
 	.probe		= sgi_probe_label,
 	.write		= sgi_write_disklabel,
 	.verify		= sgi_verify_disklabel,
+	.get_item	= sgi_get_disklabel_item,
 	.create		= sgi_create_disklabel,
-	.list		= sgi_list_table,
-	.part_add	= sgi_add_partition,
-	.part_delete	= sgi_delete_partition,
-	.part_get_type	= sgi_get_parttype,
-	.part_set_type	= sgi_set_parttype,
 
-	.part_get_status = sgi_get_partition_status,
+	.get_part	= sgi_get_partition,
+	.set_part	= sgi_set_partition,
+	.add_part	= sgi_add_partition,
+	.del_part	= sgi_delete_partition,
+
+	.part_is_used	= sgi_partition_is_used,
 	.part_toggle_flag = sgi_toggle_partition_flag
 };
 
@@ -1179,7 +1197,9 @@ struct fdisk_label *fdisk_new_sgi_label(struct fdisk_context *cxt)
 	lb->id = FDISK_DISKLABEL_SGI;
 	lb->op = &sgi_operations;
 	lb->parttypes = sgi_parttypes;
-	lb->nparttypes = ARRAY_SIZE(sgi_parttypes);
+	lb->nparttypes = ARRAY_SIZE(sgi_parttypes) - 1;
+	lb->fields = sgi_fields;
+	lb->nfields = ARRAY_SIZE(sgi_fields);
 
 	lb->flags |= FDISK_LABEL_FL_REQUIRE_GEOMETRY;
 
