@@ -36,7 +36,6 @@
 #include "c.h"
 #include "env.h"
 #include "strutils.h"
-#include "xgetpass.h"
 #include "exitcodes.h"
 #include "xalloc.h"
 #include "closestream.h"
@@ -52,7 +51,6 @@
  *  --options-source-force				MNT_OMODE_FORCE
  */
 
-static int passfd = -1;
 static int readwrite;
 
 static int mk_exit_code(struct libmnt_context *cxt, int rc);
@@ -101,35 +99,8 @@ static int table_parser_errcb(struct libmnt_table *tb __attribute__((__unused__)
 			const char *filename, int line)
 {
 	if (filename)
-		warnx(_("%s: parse error: ignore entry at line %d."),
-							filename, line);
-	return 0;
-}
-
-static char *encrypt_pass_get(struct libmnt_context *cxt)
-{
-	if (!cxt)
-		return 0;
-
-#ifdef MCL_FUTURE
-	if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
-		warn(_("couldn't lock into memory"));
-		return NULL;
-	}
-#endif
-	return xgetpass(passfd, _("Password: "));
-}
-
-static void encrypt_pass_release(struct libmnt_context *cxt
-			__attribute__((__unused__)), char *pwd)
-{
-	char *p = pwd;
-
-	while (p && *p)
-		*p++ = '\0';
-
-	free(pwd);
-	munlockall();
+		warnx(_("%s: parse error at line %d -- ignored"), filename, line);
+	return 1;
 }
 
 /*
@@ -191,7 +162,7 @@ static void print_all(struct libmnt_context *cxt, char *pattern, int show_label)
 		free(xsrc);
 	}
 
-	mnt_free_cache(cache);
+	mnt_unref_cache(cache);
 	mnt_free_iter(itr);
 }
 
@@ -225,12 +196,14 @@ static int mount_all(struct libmnt_context *cxt)
 			if (mnt_context_is_verbose(cxt))
 				printf("%-25s: mount successfully forked\n", tgt);
 		} else {
-			mk_exit_code(cxt, mntrc);	/* to print warnings */
-
-			if (mnt_context_get_status(cxt)) {
+			if (mk_exit_code(cxt, mntrc) == MOUNT_EX_SUCCESS) {
 				nsucc++;
 
-				if (mnt_context_is_verbose(cxt))
+				/* Note that MOUNT_EX_SUCCESS return code does
+				 * not mean that FS has been really mounted
+				 * (e.g. nofail option) */
+				if (mnt_context_get_status(cxt) 
+				    && mnt_context_is_verbose(cxt))
 					printf("%-25s: successfully mounted\n", tgt);
 			} else
 				nerrs++;
@@ -262,7 +235,7 @@ static int mount_all(struct libmnt_context *cxt)
 static void success_message(struct libmnt_context *cxt)
 {
 	unsigned long mflags = 0;
-	const char *tgt, *src;
+	const char *tgt, *src, *pr = program_invocation_short_name;
 
 	if (mnt_context_helper_executed(cxt)
 	    || mnt_context_get_status(cxt) != 1)
@@ -273,13 +246,16 @@ static void success_message(struct libmnt_context *cxt)
 	src = mnt_context_get_source(cxt);
 
 	if (mflags & MS_MOVE)
-		warnx(_("%s moved to %s"), src, tgt);
+		printf(_("%s: %s moved to %s.\n"), pr, src, tgt);
 	else if (mflags & MS_BIND)
-		warnx(_("%s binded on %s"), src, tgt);
-	else if (mflags & MS_PROPAGATION)
-		warnx(_("%s propagation flags changed"), tgt);
-	else
-		warnx(_("%s mounted on %s"), src, tgt);
+		printf(_("%s: %s bound on %s.\n"), pr, src, tgt);
+	else if (mflags & MS_PROPAGATION) {
+		if (src && strcmp(src, "none") != 0 && tgt)
+			printf(_("%s: %s mounted on %s.\n"), pr, src, tgt);
+
+		printf(_("%s: %s propagation flags changed.\n"), pr, tgt);
+	} else
+		printf(_("%s: %s mounted on %s.\n"), pr, src, tgt);
 }
 
 /*
@@ -346,6 +322,37 @@ static void selinux_warning(struct libmnt_context *cxt, const char *tgt)
 # define selinux_warning(_x, _y)
 #endif
 
+/*
+ * Returns 1 if @dir parent is shared
+ */
+static int is_shared_tree(struct libmnt_context *cxt, const char *dir)
+{
+	struct libmnt_table *tb = NULL;
+	struct libmnt_fs *fs;
+	unsigned long mflags = 0;
+	char *mnt = NULL, *p;
+	int rc = 0;
+
+	if (!dir)
+		return 0;
+	if (mnt_context_get_mtab(cxt, &tb) || !tb)
+		goto done;
+	mnt = xstrdup(dir);
+	p = strrchr(mnt, '/');
+	if (!p)
+		goto done;
+	if (p > mnt)
+		*p = '\0';
+	fs = mnt_table_find_mountpoint(tb, mnt, MNT_ITER_BACKWARD);
+
+	rc = fs && mnt_fs_is_kernel(fs)
+		&& mnt_fs_get_propagation(fs, &mflags) == 0
+		&& (mflags & MS_SHARED);
+done:
+	free(mnt);
+	return rc;
+}
+
 /*
  * rc = 0 success
  *     <0 error (usually -errno or -1)
@@ -363,11 +370,14 @@ static int mk_exit_code(struct libmnt_context *cxt, int rc)
 	const char *src = mnt_context_get_source(cxt);
 
 try_readonly:
-	if (mnt_context_helper_executed(cxt))
+	if (mnt_context_helper_executed(cxt)) {
 		/*
 		 * /sbin/mount.<type> called, return status
 		 */
+		if (rc == -MNT_ERR_APPLYFLAGS)
+			warnx(_("WARNING: failed to apply propagation flags"));
 		return mnt_context_get_helper_status(cxt);
+	}
 
 	if (rc == 0 && mnt_context_get_status(cxt) == 1) {
 		/*
@@ -407,6 +417,11 @@ try_readonly:
 				warnx(_("can't find mount source %s in %s"),
 						src, mnt_get_fstab_path());
 			return MOUNT_EX_USAGE;
+		case -MNT_ERR_AMBIFS:
+			warnx(_("%s: more filesystems detected. This should not happen,\n"
+			  "       use -t <type> to explicitly specify the filesystem type or\n"
+			  "       use wipefs(8) to clean up the device."), src);
+			return MOUNT_EX_USAGE;
 		case -MNT_ERR_NOFSTYPE:
 			if (restricted)
 				warnx(_("I could not determine the filesystem type, "
@@ -415,6 +430,8 @@ try_readonly:
 				warnx(_("you must specify the filesystem type"));
 			return MOUNT_EX_USAGE;
 		case -MNT_ERR_NOSOURCE:
+			if (uflags & MNT_MS_NOFAIL)
+				return MOUNT_EX_SUCCESS;
 			if (src)
 				warnx(_("can't find %s"), src);
 			else
@@ -427,13 +444,7 @@ try_readonly:
 				warnx(_("failed to parse mount options"));
 			return MOUNT_EX_USAGE;
 		case -MNT_ERR_LOOPDEV:
-			if (errno == ENOENT
-			    && (uflags & MNT_MS_ENCRYPTION)
-			    && src && stat(src, &st) == 0)
-				warnx(_("%s: failed to setup loop device "
-					"(probably unknown encryption type)"), src);
-			else
-				warn(_("%s: failed to setup loop device"), src);
+			warn(_("%s: failed to setup loop device"), src);
 			return MOUNT_EX_FAIL;
 		default:
 			return handle_generic_errors(rc, _("%s: mount failed"),
@@ -498,25 +509,30 @@ try_readonly:
 		break;
 	}
 	case ENOENT:
-		if (lstat(tgt, &st))
+		if (tgt && lstat(tgt, &st))
 			warnx(_("mount point %s does not exist"), tgt);
-		else if (stat(tgt, &st))
+		else if (tgt && stat(tgt, &st))
 			warnx(_("mount point %s is a symbolic link to nowhere"), tgt);
-		else if (stat(src, &st)) {
+		else if (src && stat(src, &st)) {
 			if (uflags & MNT_MS_NOFAIL)
 				return MOUNT_EX_SUCCESS;
 
 			warnx(_("special device %s does not exist"), src);
 		} else {
 			errno = syserr;
-			warn(_("mount(2) failed"));	/* print errno */
+			if (tgt)
+				warn("%s: %s", _("mount(2) failed"), tgt);
+			else if (src)
+				warn("%s: %s", _("mount(2) failed"), src);
+			else
+				warn(_("mount(2) failed"));
 		}
 		break;
 
 	case ENOTDIR:
 		if (stat(tgt, &st) || ! S_ISDIR(st.st_mode))
 			warnx(_("mount point %s is not a directory"), tgt);
-		else if (stat(src, &st) && errno == ENOTDIR) {
+		else if (src && stat(src, &st) && errno == ENOTDIR) {
 			if (uflags & MNT_MS_NOFAIL)
 				return MOUNT_EX_SUCCESS;
 
@@ -524,15 +540,18 @@ try_readonly:
 				 "(a path prefix is not a directory)"), src);
 		} else {
 			errno = syserr;
-			warn(_("mount(2) failed"));     /* print errno */
+			warn("%s: %s", _("mount(2) failed"), tgt);
 		}
 		break;
 
 	case EINVAL:
 		if (mflags & MS_REMOUNT)
 			warnx(_("%s not mounted or bad option"), tgt);
-		else if (mflags & MS_PROPAGATION)
+		else if (rc == -MNT_ERR_APPLYFLAGS)
 			warnx(_("%s is not mountpoint or bad option"), tgt);
+		else if ((mflags & MS_MOVE) && is_shared_tree(cxt, src))
+			warnx(_("bad option. Note that moving a mount residing under a shared\n"
+				"       mount is unsupported."));
 		else
 			warnx(_("wrong fs type, bad option, bad superblock on %s,\n"
 				"       missing codepage or helper program, or other error"),
@@ -543,9 +562,9 @@ try_readonly:
 				"       (for several filesystems (e.g. nfs, cifs) you might\n"
 				"       need a /sbin/mount.<type> helper program)\n"));
 
-		fprintf(stderr, _(
+		fprintf(stderr, _("\n"
 				"       In some cases useful info is found in syslog - try\n"
-				"       dmesg | tail or so\n"));
+				"       dmesg | tail or so.\n"));
 		break;
 
 	case EMFILE:
@@ -557,7 +576,10 @@ try_readonly:
 		break;
 
 	case ENODEV:
-		warnx(_("unknown filesystem type '%s'"), mnt_context_get_fstype(cxt));
+		if (mnt_context_get_fstype(cxt))
+			warnx(_("unknown filesystem type '%s'"), mnt_context_get_fstype(cxt));
+		else
+			warnx(_("unknown filesystem type"));
 		break;
 
 	case ENOTBLK:
@@ -593,6 +615,9 @@ try_readonly:
 		else if (mflags & MS_REMOUNT)
 			warnx(_("cannot remount %s read-write, is write-protected"), src);
 
+		else if (mflags & MS_BIND)
+			warn(_("mount %s on %s failed"), src, tgt);
+
 		else {
 			warnx(_("%s is write-protected, mounting read-only"), src);
 
@@ -607,6 +632,9 @@ try_readonly:
 		break;
 
 	case ENOMEDIUM:
+		if (uflags & MNT_MS_NOFAIL)
+			return MOUNT_EX_SUCCESS;
+
 		warnx(_("no medium found on %s"), src);
 		break;
 
@@ -630,6 +658,8 @@ static struct libmnt_table *append_fstab(struct libmnt_context *cxt,
 
 		mnt_table_set_parser_errcb(fstab, table_parser_errcb);
 		mnt_context_set_fstab(cxt, fstab);
+
+		mnt_unref_table(fstab);	/* reference is handled by @cxt now */
 	}
 
 	if (mnt_table_parse_fstab(fstab, path))
@@ -669,6 +699,24 @@ static void sanitize_paths(struct libmnt_context *cxt)
 	}
 }
 
+static void append_option(struct libmnt_context *cxt, const char *opt)
+{
+	if (opt && (*opt == '=' || *opt == '\'' || *opt == '\"' || isblank(*opt)))
+		errx(MOUNT_EX_USAGE, _("unsupported option format: %s"), opt);
+	if (mnt_context_append_options(cxt, opt))
+		err(MOUNT_EX_SYSERR, _("failed to append option '%s'"), opt);
+}
+
+static int has_remount_flag(struct libmnt_context *cxt)
+{
+	unsigned long mflags = 0;
+
+	if (mnt_context_get_mflags(cxt, &mflags))
+		return 0;
+
+	return mflags & MS_REMOUNT;
+}
+
 static void __attribute__((__noreturn__)) usage(FILE *out)
 {
 	fputs(USAGE_HEADER, out);
@@ -680,6 +728,9 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 		" %1$s <operation> <mountpoint> [<target>]\n"),
 		program_invocation_short_name);
 
+	fputs(USAGE_SEPARATOR, out);
+	fputs(_("Mount a filesystem.\n"), out);
+
 	fputs(USAGE_OPTIONS, out);
 	fprintf(out, _(
 	" -a, --all               mount all filesystems mentioned in fstab\n"
@@ -688,23 +739,23 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	" -F, --fork              fork off for each device (use with -a)\n"
 	" -T, --fstab <path>      alternative file to /etc/fstab\n"));
 	fprintf(out, _(
-	" -h, --help              display this help text and exit\n"
-	" -i, --internal-only     don't call the mount.<type> helpers\n"
-	" -l, --show-labels       lists all mounts with LABELs\n"
+	" -i, --internal-only     don't call the mount.<type> helpers\n"));
+	fprintf(out, _(
+	" -l, --show-labels       show also filesystem labels\n"));
+	fprintf(out, _(
 	" -n, --no-mtab           don't write to /etc/mtab\n"));
 	fprintf(out, _(
 	" -o, --options <list>    comma-separated list of mount options\n"
 	" -O, --test-opts <list>  limit the set of filesystems (use with -a)\n"
-	" -p, --pass-fd <num>     read the passphrase from file descriptor\n"
 	" -r, --read-only         mount the filesystem read-only (same as -o ro)\n"
 	" -t, --types <list>      limit the set of filesystem types\n"));
 	fprintf(out, _(
 	"     --source <src>      explicitly specifies source (path, label, uuid)\n"
 	"     --target <target>   explicitly specifies mountpoint\n"));
 	fprintf(out, _(
-	" -v, --verbose           say what is being done\n"
-	" -V, --version           display version information and exit\n"
-	" -w, --read-write        mount the filesystem read-write (default)\n"));
+	" -v, --verbose           say what is being done\n"));
+	fprintf(out, _(
+	" -w, --rw, --read-write  mount the filesystem read-write (default)\n"));
 
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
@@ -753,6 +804,7 @@ int main(int argc, char **argv)
 	char *srcbuf = NULL;
 	char *types = NULL;
 	unsigned long oper = 0;
+	int propa = 0;
 
 	enum {
 		MOUNT_OPT_SHARED = CHAR_MAX + 1,
@@ -782,7 +834,6 @@ int main(int argc, char **argv)
 		{ "rw", 0, 0, 'w' },
 		{ "options", 1, 0, 'o' },
 		{ "test-opts", 1, 0, 'O' },
-		{ "pass-fd", 1, 0, 'p' },
 		{ "types", 1, 0, 't' },
 		{ "uuid", 1, 0, 'U' },
 		{ "label", 1, 0, 'L'},
@@ -806,12 +857,7 @@ int main(int argc, char **argv)
 	};
 
 	static const ul_excl_t excl[] = {       /* rows and cols in in ASCII order */
-		{ 'B','M','R',			/* bind,move,rbind */
-		   MOUNT_OPT_SHARED,   MOUNT_OPT_SLAVE,
-		   MOUNT_OPT_PRIVATE,  MOUNT_OPT_UNBINDABLE,
-		   MOUNT_OPT_RSHARED,  MOUNT_OPT_RSLAVE,
-		   MOUNT_OPT_RPRIVATE, MOUNT_OPT_RUNBINDABLE },
-
+		{ 'B','M','R' },			/* bind,move,rbind */
 		{ 'L','U', MOUNT_OPT_SOURCE },	/* label,uuid,source */
 		{ 0 }
 	};
@@ -830,12 +876,12 @@ int main(int argc, char **argv)
 
 	mnt_context_set_tables_errcb(cxt, table_parser_errcb);
 
-	while ((c = getopt_long(argc, argv, "aBcfFhilL:Mno:O:p:rRsU:vVwt:T:",
+	while ((c = getopt_long(argc, argv, "aBcfFhilL:Mno:O:rRsU:vVwt:T:",
 					longopts, NULL)) != -1) {
 
 		/* only few options are allowed for non-root users */
 		if (mnt_context_is_restricted(cxt) &&
-		    !strchr("hlLUVvpris", c) &&
+		    !strchr("hlLUVvrist", c) &&
 		    c != MOUNT_OPT_TARGET &&
 		    c != MOUNT_OPT_SOURCE)
 			exit_non_root(option_to_longopt(c, longopts));
@@ -865,8 +911,7 @@ int main(int argc, char **argv)
 			mnt_context_disable_mtab(cxt, TRUE);
 			break;
 		case 'r':
-			if (mnt_context_append_options(cxt, "ro"))
-				err(MOUNT_EX_SYSERR, _("failed to append options"));
+			append_option(cxt, "ro");
 			readwrite = 0;
 			break;
 		case 'v':
@@ -876,22 +921,16 @@ int main(int argc, char **argv)
 			print_version();
 			break;
 		case 'w':
-			if (mnt_context_append_options(cxt, "rw"))
-				err(MOUNT_EX_SYSERR, _("failed to append options"));
+			append_option(cxt, "rw");
 			readwrite = 1;
 			break;
 		case 'o':
-			if (mnt_context_append_options(cxt, optarg))
-				err(MOUNT_EX_SYSERR, _("failed to append options"));
+			append_option(cxt, optarg);
 			break;
 		case 'O':
 			if (mnt_context_set_options_pattern(cxt, optarg))
 				err(MOUNT_EX_SYSERR, _("failed to set options pattern"));
 			break;
-		case 'p':
-			passfd = strtou32_or_err(optarg,
-					_("invalid passphrase file descriptor"));
-			break;
 		case 'L':
 			xasprintf(&srcbuf, "LABEL=\"%s\"", optarg);
 			mnt_context_disable_swapmatch(cxt, 1);
@@ -926,28 +965,36 @@ int main(int argc, char **argv)
 			oper |= (MS_BIND | MS_REC);
 			break;
 		case MOUNT_OPT_SHARED:
-			oper |= MS_SHARED;
+			append_option(cxt, "shared");
+			propa = 1;
 			break;
 		case MOUNT_OPT_SLAVE:
-			oper |= MS_SLAVE;
+			append_option(cxt, "slave");
+			propa = 1;
 			break;
 		case MOUNT_OPT_PRIVATE:
-			oper |= MS_PRIVATE;
+			append_option(cxt, "private");
+			propa = 1;
 			break;
 		case MOUNT_OPT_UNBINDABLE:
-			oper |= MS_UNBINDABLE;
+			append_option(cxt, "unbindable");
+			propa = 1;
 			break;
 		case MOUNT_OPT_RSHARED:
-			oper |= (MS_SHARED | MS_REC);
+			append_option(cxt, "rshared");
+			propa = 1;
 			break;
 		case MOUNT_OPT_RSLAVE:
-			oper |= (MS_SLAVE | MS_REC);
+			append_option(cxt, "rslave");
+			propa = 1;
 			break;
 		case MOUNT_OPT_RPRIVATE:
-			oper |= (MS_PRIVATE | MS_REC);
+			append_option(cxt, "rprivate");
+			propa = 1;
 			break;
 		case MOUNT_OPT_RUNBINDABLE:
-			oper |= (MS_UNBINDABLE | MS_REC);
+			append_option(cxt, "runbindable");
+			propa = 1;
 			break;
 		case MOUNT_OPT_TARGET:
 			mnt_context_disable_swapmatch(cxt, 1);
@@ -966,16 +1013,30 @@ int main(int argc, char **argv)
 	argc -= optind;
 	argv += optind;
 
+	if (fstab && !mnt_context_is_nocanonicalize(cxt)) {
+		/*
+		 * We have external (context independent) fstab instance, let's
+		 * make a connection between the fstab and the canonicalization
+		 * cache.
+		 */
+		mnt_table_set_cache(fstab, mnt_context_get_cache(cxt));
+	}
+
 	if (!mnt_context_get_source(cxt) &&
 	    !mnt_context_get_target(cxt) &&
 	    !argc &&
 	    !all) {
-		if (oper)
+		if (oper || mnt_context_get_options(cxt))
 			usage(stderr);
 		print_all(cxt, types, show_labels);
 		goto done;
 	}
 
+	/* Non-root users are allowed to use -t to print_all(),
+	   but not to mount */
+	if (mnt_context_is_restricted(cxt) && types)
+		exit_non_root("types");
+
 	if (oper && (types || all || mnt_context_get_source(cxt)))
 		usage(stderr);
 
@@ -985,8 +1046,6 @@ int main(int argc, char **argv)
 	else if (types)
 		mnt_context_set_fstype(cxt, types);
 
-	mnt_context_set_passwd_cb(cxt, encrypt_pass_get, encrypt_pass_release);
-
 	if (all) {
 		/*
 		 * A) Mount all
@@ -998,24 +1057,41 @@ int main(int argc, char **argv)
 				 mnt_context_get_target(cxt))) {
 		/*
 		 * B) mount -L|-U|--source|--target
+		 *
+		 * non-root may specify source *or* target, but not both
 		 */
 		if (mnt_context_is_restricted(cxt) &&
 		    mnt_context_get_source(cxt) &&
 		    mnt_context_get_target(cxt))
 			exit_non_root(NULL);
 
-	} else if (argc == 1) {
+	} else if (argc == 1 && (!mnt_context_get_source(cxt) ||
+				 !mnt_context_get_target(cxt))) {
 		/*
 		 * C) mount [-L|-U|--source] <target>
+		 *    mount [--target <dir>] <source>
 		 *    mount <source|target>
 		 *
 		 * non-root may specify source *or* target, but not both
+		 *
+		 * It does not matter for libmount if we set source or target
+		 * here (the library is able to swap it), but it matters for
+		 * sanitize_paths().
 		 */
-		if (mnt_context_is_restricted(cxt) &&
-		    mnt_context_get_source(cxt))
-			exit_non_root(NULL);
+		int istag = mnt_tag_is_valid(argv[0]);
 
-		mnt_context_set_target(cxt, argv[0]);
+		if (istag && mnt_context_get_source(cxt))
+			/* -L, -U or --source together with LABEL= or UUID= */
+			errx(MOUNT_EX_USAGE, _("source specified more than once"));
+		else if (istag || mnt_context_get_target(cxt))
+			mnt_context_set_source(cxt, argv[0]);
+		else
+			mnt_context_set_target(cxt, argv[0]);
+
+		if (mnt_context_is_restricted(cxt) &&
+		    mnt_context_get_source(cxt) &&
+		    mnt_context_get_target(cxt))
+			exit_non_root(NULL);
 
 	} else if (argc == 2 && !mnt_context_get_source(cxt)
 			     && !mnt_context_get_target(cxt)) {
@@ -1024,6 +1100,7 @@ int main(int argc, char **argv)
 		 */
 		if (mnt_context_is_restricted(cxt))
 			exit_non_root(NULL);
+
 		mnt_context_set_source(cxt, argv[0]);
 		mnt_context_set_target(cxt, argv[1]);
 
@@ -1033,13 +1110,13 @@ int main(int argc, char **argv)
 	if (mnt_context_is_restricted(cxt))
 		sanitize_paths(cxt);
 
-	if (oper) {
-		/* MS_PROPAGATION operations, let's set the mount flags */
+	if (oper)
+		/* BIND/MOVE operations, let's set the mount flags */
 		mnt_context_set_mflags(cxt, oper);
 
-		/* For -make* or --bind is fstab unnecessary */
+	if ((oper && !has_remount_flag(cxt)) || propa)
+		/* For --make-* or --bind is fstab/mtab unnecessary */
 		mnt_context_set_optsmode(cxt, MNT_OMODE_NOTAB);
-	}
 
 	rc = mnt_context_mount(cxt);
 	rc = mk_exit_code(cxt, rc);
@@ -1048,7 +1125,6 @@ int main(int argc, char **argv)
 		success_message(cxt);
 done:
 	mnt_free_context(cxt);
-	mnt_free_table(fstab);
 	return rc;
 }
 
